"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018 Onur Rauf Bingol

    Tests geomdl.GPGen module. Requires "pytest" to run.
"""

import pytest
from geomdl import CPGen

GRID_TOL = 10e-8


@pytest.fixture
def grid():
    """ Generates a 3x4 control points grid """
    surfgrid = CPGen.Grid(7, 13)
    surfgrid.generate(3, 4)
    return surfgrid


@pytest.fixture
def grid2():
    """ Generates a 6x6 control points grid """
    surfgrid = CPGen.Grid(7, 13)
    surfgrid.generate(16, 16)
    return surfgrid


@pytest.fixture
def gridw():
    """ Generates a weighted control points grid """
    surfgrid = CPGen.GridWeighted(7, 13)
    surfgrid.generate(3, 4)
    return surfgrid


def test_grid_generate1():
    test_grid = CPGen.GridWeighted(7, 13)
    with pytest.raises(ValueError):
        test_grid.generate(-1, 5)


def test_grid_generate2():
    test_grid = CPGen.GridWeighted(7, 13)
    with pytest.raises(ValueError):
        test_grid.generate(5, -1)


def test_grid_generate3():
    test_grid = CPGen.GridWeighted(7, 13)
    with pytest.warns(UserWarning):
        test_grid.generate(3.5, 4)


def test_grid_generate4():
    test_grid = CPGen.GridWeighted(7, 13)
    with pytest.warns(UserWarning):
        test_grid.generate(3, 4.2)


def test_grid_generate5(grid):
    # testing regeneration of the grid
    grid.generate(13, 17)

    assert grid._size_u == 13
    assert grid._size_v == 17
    assert len(grid.grid) == 14
    assert len(grid.grid[0]) == 18


def test_grid(grid):
    result = [[[0.0, 0.0, 0.0], [0.0, 3.25, 0.0], [0.0, 6.5, 0.0], [0.0, 9.75, 0.0], [0.0, 13.0, 0.0]],
              [[2.3333333333333335, 0.0, 0.0], [2.3333333333333335, 3.25, 0.0],
               [2.3333333333333335, 6.5, 0.0], [2.3333333333333335, 9.75, 0.0], [2.3333333333333335, 13.0, 0.0]],
              [[4.666666666666667, 0.0, 0.0], [4.666666666666667, 3.25, 0.0], [4.666666666666667, 6.5, 0.0],
               [4.666666666666667, 9.75, 0.0], [4.666666666666667, 13.0, 0.0]],
              [[7.0, 0.0, 0.0], [7.0, 3.25, 0.0], [7.0, 6.5, 0.0], [7.0, 9.75, 0.0], [7.0, 13.0, 0.0]]]

    assert grid.grid == result


def test_grid_reset1(grid):
    grid.reset()

    assert grid._grid_points == []
    assert grid._size_u == 0
    assert grid._size_v == 0
    assert grid._origin == [0.0, 0.0, 0.0]


def test_grid_reset2(gridw):
    gridw.weight = 0.33
    gridw.reset()

    assert gridw._grid_points == []
    assert gridw._size_u == 0
    assert gridw._size_v == 0
    assert gridw._origin == [0.0, 0.0, 0.0]
    assert gridw._cache['gridptsw'] == []
    assert gridw._weights == []


def test_grid_weight1(gridw):
    result = [[[0.0, 0.0, 0.0, 1.0], [0.0, 3.25, 0.0, 1.0], [0.0, 6.5, 0.0, 1.0],
               [0.0, 9.75, 0.0, 1.0], [0.0, 13.0, 0.0, 1.0]],
              [[2.3333333333333335, 0.0, 0.0, 1.0], [2.3333333333333335, 3.25, 0.0, 1.0],
               [2.3333333333333335, 6.5, 0.0, 1.0], [2.3333333333333335, 9.75, 0.0, 1.0],
               [2.3333333333333335, 13.0, 0.0, 1.0]],
              [[4.666666666666667, 0.0, 0.0, 1.0],
               [4.666666666666667, 3.25, 0.0, 1.0], [4.666666666666667, 6.5, 0.0, 1.0],
               [4.666666666666667, 9.75, 0.0, 1.0], [4.666666666666667, 13.0, 0.0, 1.0]],
              [[7.0, 0.0, 0.0, 1.0], [7.0, 3.25, 0.0, 1.0], [7.0, 6.5, 0.0, 1.0],
               [7.0, 9.75, 0.0, 1.0], [7.0, 13.0, 0.0, 1.0]]]

    # default weight is 1.0
    assert gridw.grid == result


def test_grid_weight2(gridw):
    with pytest.raises(TypeError):
        gridw.weight = "test string"


def test_grid_weight3(gridw):
    with pytest.raises(ValueError):
        gridw.weight = -0.1


def test_grid_weight4(gridw):
    # default weights vector should be empty
    assert gridw._weights == []


def test_grid_weight5(gridw):
    # try to change weight
    gridw.weight = 0.35
    assert gridw.weight == [0.35 for _ in range(len(gridw))]


def test_grid_weight6(gridw):
    result = [[[0.0, 0.0, 0.0, 1.0], [0.0, 3.25, 0.0, 1.0], [0.0, 6.5, 0.0, 1.0],
               [0.0, 9.75, 0.0, 1.0], [0.0, 13.0, 0.0, 1.0]],
              [[2.3333333333333335, 0.0, 0.0, 1.0], [2.3333333333333335, 3.25, 0.0, 1.0],
               [2.3333333333333335, 6.5, 0.0, 1.0], [2.3333333333333335, 9.75, 0.0, 1.0],
               [2.3333333333333335, 13.0, 0.0, 1.0]],
              [[4.666666666666667, 0.0, 0.0, 1.0],
               [4.666666666666667, 3.25, 0.0, 1.0], [4.666666666666667, 6.5, 0.0, 1.0],
               [4.666666666666667, 9.75, 0.0, 1.0], [4.666666666666667, 13.0, 0.0, 1.0]],
              [[7.0, 0.0, 0.0, 1.0], [7.0, 3.25, 0.0, 1.0], [7.0, 6.5, 0.0, 1.0],
               [7.0, 9.75, 0.0, 1.0], [7.0, 13.0, 0.0, 1.0]]]

    gen_cache_var = gridw.grid

    # this should pull up the grid from the cache
    assert gridw.grid == result


def test_bumps1(grid2):
    grid2.bumps(num_bumps=2, bump_height=[5.0, 7.0], base_extent=2)
    check_vals = grid2.grid

    check = False
    for rows in check_vals:
        for val in rows:
            if val[2] == 5.0 or val[2] == 7.0:
                check = True

    assert check


def test_bumps1_all_positive_heights(grid2):
    grid2.bumps(num_bumps=2, bump_height=5, base_extent=2)
    check_vals = grid2.grid

    check = False
    for rows in check_vals:
        for val in rows:
            if val[2] == 5.0:
                check = True

    assert check


def test_bumps2(grid2):
    with pytest.raises(RuntimeError):
        # impossible to add 100 bumps on this specific grid
        grid2.bumps(num_bumps=100, bump_height=5, base_extent=2)


def test_bumps4():
    test_grid = CPGen.Grid(5, 7)
    with pytest.raises(RuntimeError):
        # bumps before calling generate()
        test_grid.bumps(num_bumps=3)


def test_bumps5(grid2):
    with pytest.warns(UserWarning):
        # non-integer num_bumps
        grid2.bumps(num_bumps=1.1, bump_height=5, base_extent=2)


def test_bumps8(grid2):
    with pytest.raises(ValueError):
        # large base_extent
        grid2.bumps(num_bumps=1, bump_height=5, base_extent=20)


def test_bumps9(grid2):
    with pytest.raises(ValueError):
        # small base_extent
        grid2.bumps(num_bumps=1, bump_height=5, base_extent=0)



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018-2019 Onur Rauf Bingol

    Requires "pytest" to run.
"""

from geomdl import helpers

GEOMDL_DELTA = 10e-8


def test_find_span_binsearch():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	num_ctrlpts = len(knot_vector) - degree - 1
	knot = 5.0 / 2.0

	to_check = helpers.find_span_binsearch(degree, knot_vector, num_ctrlpts, knot)
	result = 4  # Value from The Nurbs Book p.68

	assert to_check == result


def test_basis_function():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	span = 4
	knot = 5.0 / 2.0

	to_check = helpers.basis_function(degree, knot_vector, span, knot)
	result = [1.0 / 8.0, 6.0 / 8.0, 1.0 / 8.0]  # Values from The Nurbs Book p.69

	assert abs(to_check[0] - result[0]) < GEOMDL_DELTA
	assert abs(to_check[1] - result[1]) < GEOMDL_DELTA
	assert abs(to_check[2] - result[2]) < GEOMDL_DELTA


def test_basis_functions():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	spans = [3, 4, 5]
	knots = [5.0 / 2.0 for _ in range(len(spans))]

	to_check = helpers.basis_functions(degree, knot_vector, spans, knots)
	result = [helpers.basis_function(degree, knot_vector, spans[_], knots[_]) for _ in range(len(spans))]  # Values from The Nurbs Book p.55

	assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
	assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
	assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA

	assert abs(to_check[1][0] - result[1][0]) < GEOMDL_DELTA
	assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
	assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA

	assert abs(to_check[2][0] - result[2][0]) < GEOMDL_DELTA
	assert abs(to_check[2][1] - result[2][1]) < GEOMDL_DELTA
	assert abs(to_check[2][2] - result[2][2]) < GEOMDL_DELTA


def test_basis_function_all():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	span = 4
	knot = 5.0 / 2.0

	to_check = helpers.basis_function_all(degree, knot_vector, span, knot)
	interm = [helpers.basis_function(_, knot_vector, span, knot) + [None] * (degree - _) for _ in range(0, degree + 1)]
	result = [list(_) for _ in zip(*interm)]  # Transposing to the same format as to_check

	assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
	assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
	assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA

	assert to_check[1][0] == result[1][0]  # NoneType can't be subtracted
	assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
	assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA

	assert to_check[2][0] == result[2][0]  # None
	assert to_check[2][1] == result[2][1]  # None
	assert abs(to_check[2][2] - result[2][2]) < GEOMDL_DELTA


def test_basis_function_ders():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	span = 4
	knot = 5.0 / 2.0
	order = 2

	to_check = helpers.basis_function_ders(degree, knot_vector, span, knot, order)
	result = [[0.125, 0.75, 0.125], [-0.5, 0.0, 0.5], [1.0, -2.0, 1.0]] # Values and formulas from The Nurbs Book p.69 & p.72

	assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
	assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
	assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA

	assert abs(to_check[1][0] - result[1][0]) < GEOMDL_DELTA
	assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
	assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA

	assert abs(to_check[2][0] - result[2][0]) < GEOMDL_DELTA
	assert abs(to_check[2][1] - result[2][1]) < GEOMDL_DELTA
	assert abs(to_check[2][2] - result[2][2]) < GEOMDL_DELTA


def test_find_multiplicity():
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	knots = [0.5, 2, 4, 5]

	to_check = [helpers.find_multiplicity(_, knot_vector) for _ in knots]
	result = [0, 1, 2, 3]

	assert to_check == result


def test_basis_function_one():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	span = 4
	knot = 5.0 / 2.0
	
	# Values from basis_function
	to_check = [helpers.basis_function_one(degree, knot_vector, span - _, knot) for _ in range(degree, -1, -1)]
	result = helpers.basis_function(degree, knot_vector, span, knot)

	assert abs(to_check[0] - result[0]) < GEOMDL_DELTA
	assert abs(to_check[1] - result[1]) < GEOMDL_DELTA
	assert abs(to_check[2] - result[2]) < GEOMDL_DELTA


def test_basis_function_ders_one():
	degree = 2
	knot_vector = [0, 0, 0, 1, 2, 3, 4, 4, 5, 5, 5]
	span = 4
	knot = 5.0 / 2.0
	order = 2

	to_check = [helpers.basis_function_ders_one(degree, knot_vector, span - _, knot, order) for _ in range(degree, -1, -1)]
	interm = helpers.basis_function_ders(degree, knot_vector, span, knot, order)
	result = [list(_) for _ in zip(*interm)]

	assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
	assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
	assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA

	assert abs(to_check[1][0] - result[1][0]) < GEOMDL_DELTA
	assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
	assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA

	assert abs(to_check[2][0] - result[2][0]) < GEOMDL_DELTA
	assert abs(to_check[2][1] - result[2][1]) < GEOMDL_DELTA
	assert abs(to_check[2][2] - result[2][2]) < GEOMDL_DELTA



"""
.. module:: voxelize
    :platform: Unix, Windows
    :synopsis: Provides voxelization functions

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

import struct
from . import _voxelize as vxl
from ._utilities import export


@export
def voxelize(obj, **kwargs):
    """ Generates binary voxel representation of the surfaces and volumes.

    Keyword Arguments:
        * ``grid_size``: size of the voxel grid. *Default: (8, 8, 8)*
        * ``padding``: voxel padding for in-outs finding. *Default: 10e-8*
        * ``use_cubes``: use cube voxels instead of cuboid ones. *Default: False*
        * ``num_procs``: number of concurrent processes for voxelization. *Default: 1*

    :param obj: input surface(s) or volume(s)
    :type obj: abstract.Surface or abstract.Volume
    :return: voxel grid and filled information
    :rtype: tuple
    """
    # Get keyword arguments
    grid_size = kwargs.pop('grid_size', (8, 8, 8))
    use_cubes = kwargs.pop('use_cubes', False)
    num_procs = kwargs.get('num_procs', 1)

    if not isinstance(grid_size, (list, tuple)):
        raise TypeError("Grid size must be a list or a tuple of integers")

    # Initialize result arrays
    grid = []
    filled = []

    # Should also work with multi surfaces and volumes
    for o in obj:
        # Generate voxel grid
        grid_temp = vxl.generate_voxel_grid(o.bbox, grid_size, use_cubes=use_cubes)
        args = [grid_temp, o.evalpts]

        # Find in-outs
        filled_temp = vxl.find_inouts_mp(*args, **kwargs) if num_procs > 1 else vxl.find_inouts_st(*args, **kwargs)

        # Add to result arrays
        grid += grid_temp
        filled += filled_temp

    # Return result arrays
    return grid, filled


def convert_bb_to_faces(voxel_grid):
    """ Converts a voxel grid defined by min and max coordinates to a voxel grid defined by faces.

    :param voxel_grid: voxel grid defined by the bounding box of all voxels
    :return: voxel grid with face data
    """
    new_vg = []
    for v in voxel_grid:
        # Vertices
        p1 = v[0]
        p2 = [v[1][0], v[0][1], v[0][2]]
        p3 = [v[1][0], v[1][1], v[0][2]]
        p4 = [v[0][0], v[1][1], v[0][2]]
        p5 = [v[0][0], v[0][1], v[1][2]]
        p6 = [v[1][0], v[0][1], v[1][2]]
        p7 = v[1]
        p8 = [v[0][0], v[1][1], v[1][2]]
        # Faces
        fb = [p1, p2, p3, p4]  # bottom face
        ft = [p5, p6, p7, p8]  # top face
        fs1 = [p1, p2, p6, p5]  # side face 1
        fs2 = [p2, p3, p7, p6]  # side face 2
        fs3 = [p3, p4, p8, p7]  # side face 3
        fs4 = [p1, p4, p8, p5]  # side face 4
        # Append to return list
        new_vg.append([fb, fs1, fs2, fs3, fs4, ft])
    return new_vg


@export
def save_voxel_grid(voxel_grid, file_name):
    """ Saves binary voxel grid as a binary file.

    The binary file is structured in little-endian unsigned int format.

    :param voxel_grid: binary voxel grid
    :type voxel_grid: list, tuple
    :param file_name: file name to save
    :type file_name: str
    """
    try:
        with open(file_name, 'wb') as fp:
            for voxel in voxel_grid:
                fp.write(struct.pack("<I", voxel))
    except IOError as e:
        print("An error occurred: {}".format(e.args[-1]))
        raise e
    except Exception:
        raise



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018-2019 Onur Rauf Bingol

    Requires "pytest" to run.
"""

from pytest import fixture, mark
from geomdl import BSpline
from geomdl import evaluators
from geomdl import helpers
from geomdl import convert
from geomdl import operations
from geomdl import linalg

GEOMDL_DELTA = 0.001


@fixture
def spline_curve():
    """ Creates a spline Curve """
    curve = BSpline.Curve()
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]
    return curve


def test_bspline_curve_name(spline_curve):
    spline_curve.name = "Curve Testing"
    assert spline_curve.name == "Curve Testing"


def test_bspline_curve_degree(spline_curve):
    assert spline_curve.degree == 3


def test_bspline_curve_ctrlpts(spline_curve):
    assert spline_curve.ctrlpts == [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    assert spline_curve.dimension == 2


def test_bspline_curve_knot_vector(spline_curve):
    assert spline_curve.knotvector == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


@mark.parametrize("param, res", [
    (0.0, (5.0, 5.0)),
    (0.3, (18.617, 13.377)),
    (0.5, (27.645, 14.691)),
    (0.6, (32.143, 14.328)),
    (1.0, (50.0, 5.0))
])
def test_bspline_curve2d_eval(spline_curve, param, res):
    evalpt = spline_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


def test_bspline_curve2d_deriv(spline_curve):
    der1 = spline_curve.derivatives(u=0.35, order=2)
    spline_curve.evaluator = evaluators.CurveEvaluator2()
    der2 = spline_curve.derivatives(u=0.35, order=2)

    assert abs(der1[0][0] - der2[0][0]) < GEOMDL_DELTA
    assert abs(der1[0][1] - der2[0][1]) < GEOMDL_DELTA
    assert abs(der1[1][0] - der2[1][0]) < GEOMDL_DELTA
    assert abs(der1[1][1] - der2[1][1]) < GEOMDL_DELTA
    assert abs(der1[2][0] - der2[2][0]) < GEOMDL_DELTA
    assert abs(der1[2][1] - der2[2][1]) < GEOMDL_DELTA


def test_bspline_curve2d_deriv_eval(spline_curve):
    evalpt = spline_curve.evaluate_single(0.35)
    der1 = spline_curve.derivatives(u=0.35)
    spline_curve.evaluator = evaluators.CurveEvaluator2()
    der2 = spline_curve.derivatives(u=0.35)

    assert abs(der1[0][0] - evalpt[0]) < GEOMDL_DELTA
    assert abs(der1[0][1] - evalpt[1]) < GEOMDL_DELTA
    assert abs(der2[0][0] - evalpt[0]) < GEOMDL_DELTA
    assert abs(der2[0][1] - evalpt[1]) < GEOMDL_DELTA


@mark.parametrize("param, num_insert, res", [
    (0.3, 1, (18.617, 13.377)),
    (0.6, 1, (32.143, 14.328)),
    (0.6, 2, (32.143, 14.328))
])
def test_bspline_curve2d_insert_knot(spline_curve, param, num_insert, res):
    s_pre = helpers.find_multiplicity(param, spline_curve.knotvector)
    spline_curve.insert_knot(param, num=num_insert)
    s_post = helpers.find_multiplicity(param, spline_curve.knotvector)
    evalpt = spline_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert s_pre + num_insert == s_post


def test_bspline_curve2d_insert_knot_kv(spline_curve):
    spline_curve.insert_knot(0.66, num=2)
    s = helpers.find_multiplicity(0.66, spline_curve.knotvector)

    assert spline_curve.knotvector[5] == 0.66
    assert s == 3


@mark.parametrize("param, num_remove", [
    (0.33, 1),
    (0.66, 1)
])
def test_bspline_curve2d_remove_knot(spline_curve, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_curve.knotvector)
    c_pre = spline_curve.ctrlpts_size
    spline_curve.remove_knot(param, num=num_remove)
    s_post = helpers.find_multiplicity(param, spline_curve.knotvector)
    c_post = spline_curve.ctrlpts_size

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


def test_bspline_curve2d_remove_knot_kv(spline_curve):
    spline_curve.remove_knot(0.66, num=1)
    s = helpers.find_multiplicity(0.66, spline_curve.knotvector)

    assert 0.66 not in spline_curve.knotvector
    assert s == 0


@mark.parametrize("density, kv", [
    (0, [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]),
    (1, [0.0, 0.0, 0.0, 0.0, 0.165, 0.165, 0.165, 0.33, 0.33, 0.33, 0.495, 0.495, 0.495, 0.66, 0.66, 0.66, 0.830, 0.830, 0.830, 1.0, 1.0, 1.0, 1.0]),
])
def test_bspline_curve2d_knot_refine(spline_curve, density, kv):
    operations.refine_knotvector(spline_curve, [density])
    for a, b in zip(kv, spline_curve.knotvector):
        assert abs(a - b) < GEOMDL_DELTA


def test_bspline_curve2d_degree_elevate_degree(spline_curve):
    dops = 1
    degree_new = spline_curve.degree + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.degree == degree_new


def test_bspline_curve2d_degree_elevate_ctrlpts_size(spline_curve):
    dops = 1
    ctrlpts_size = spline_curve.ctrlpts_size + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.ctrlpts_size == ctrlpts_size


def test_bspline_curve2d_degree_reduce_degree(spline_curve):
    dops = -1
    degree_new = spline_curve.degree + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.degree == degree_new


def test_bspline_curve2d_degree_reduce_ctrlpts_size(spline_curve):
    dops = -1
    ctrlpts_size_new = spline_curve.ctrlpts_size + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.ctrlpts_size == ctrlpts_size_new


@fixture
def spline_curve3d(spline_curve):
    curve3d = operations.add_dimension(spline_curve, offset=1.0)
    return curve3d


def test_bspline_curve3d_ctrlpts(spline_curve3d):
    assert spline_curve3d.ctrlpts == [[5.0, 5.0, 1.0], [10.0, 10.0, 1.0], [20.0, 15.0, 1.0], [35.0, 15.0, 1.0], [45.0, 10.0, 1.0], [50.0, 5.0, 1.0]]
    assert spline_curve3d.dimension == 3


@fixture
def nurbs_curve(spline_curve):
    curve = convert.bspline_to_nurbs(spline_curve)
    curve.weights = [0.5, 1.0, 0.75, 1.0, 0.25, 1.0]
    return curve


def test_nurbs_curve2d_weights(nurbs_curve):
    assert nurbs_curve.weights == [0.5, 1.0, 0.75, 1.0, 0.25, 1.0]


@mark.parametrize("param, res", [
    (0.0, (5.0, 5.0)),
    (0.2, (13.8181, 11.5103)),
    (0.5, (28.1775, 14.7858)),
    (0.95, (48.7837, 6.0022))
])
def test_nurbs_curve2d_eval(nurbs_curve, param, res):
    evalpt = nurbs_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


@mark.parametrize("param, order, res", [
    (0.0, 1, ((5.0, 5.0), (90.9090, 90.9090))),
    (0.2, 2, ((13.8181, 11.5103), (40.0602, 17.3878), (104.4062, -29.3672))),
    (0.5, 3, ((28.1775, 14.7858), (39.7272, 2.2562), (-116.9254, -49.7367), (125.5276, 196.8865))),
    (0.95, 1, ((48.7837, 6.0022), (39.5178, -29.9962)))
])
def test_nurbs_curve2d_deriv(nurbs_curve, param, order, res):
    deriv = nurbs_curve.derivatives(u=param, order=order)

    for computed, expected in zip(deriv, res):
        for c, e in zip(computed, expected):
            assert abs(c - e) < GEOMDL_DELTA


@fixture
def spline_curve_kv_norm1():
    """ Creates a spline Curve with knot vector normalization """
    curve = BSpline.Curve(normalize_kv=True)
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0]
    return curve


@mark.parametrize("param, res", [
    (0.0, (5.0, 5.0)),
    (0.33333, (19.9998, 13.7499)),
    (0.5, (27.5, 14.687)),
    (0.66666, (34.9997, 13.75)),
    (1.0, (50.0, 5.0))
])
def test_bspline_curve2d_eval_kv_norm1(spline_curve_kv_norm1, param, res):
    evalpt = spline_curve_kv_norm1.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


@fixture
def spline_curve_kv_norm2():
    """ Creates a spline Curve without knot vector normalization """
    curve = BSpline.Curve(normalize_kv=False)
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0]
    return curve


@mark.parametrize("param, res", [
    (0.0, (5.0, 5.0)),
    (1.0, (19.9998, 13.7499)),
    (1.5, (27.5, 14.687)),
    (2.0, (34.9997, 13.75)),
    (3.0, (50.0, 5.0))
])
def test_bspline_curve2d_eval_kv_norm2(spline_curve_kv_norm2, param, res):
    evalpt = spline_curve_kv_norm2.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018-2019 Onur Rauf Bingol

    Requires "pytest" to run.
"""

from pytest import fixture, mark
from geomdl import BSpline
from geomdl import evaluators
from geomdl import convert
from geomdl import helpers

GEOMDL_DELTA = 0.001


@fixture
def spline_surf():
    """ Creates a B-spline surface instance """
    # Create a surface instance
    surf = BSpline.Surface()

    # Set degrees
    surf.degree_u = 3
    surf.degree_v = 3

    ctrlpts = [[-25.0, -25.0, -10.0], [-25.0, -15.0, -5.0], [-25.0, -5.0, 0.0], [-25.0, 5.0, 0.0],
               [-25.0, 15.0, -5.0], [-25.0, 25.0, -10.0], [-15.0, -25.0, -8.0], [-15.0, -15.0, -4.0],
               [-15.0, -5.0, -4.0], [-15.0, 5.0, -4.0], [-15.0, 15.0, -4.0], [-15.0, 25.0, -8.0],
               [-5.0, -25.0, -5.0], [-5.0, -15.0, -3.0], [-5.0, -5.0, -8.0], [-5.0, 5.0, -8.0],
               [-5.0, 15.0, -3.0], [-5.0, 25.0, -5.0], [5.0, -25.0, -3.0], [5.0, -15.0, -2.0],
               [5.0, -5.0, -8.0], [5.0, 5.0, -8.0], [5.0, 15.0, -2.0], [5.0, 25.0, -3.0],
               [15.0, -25.0, -8.0], [15.0, -15.0, -4.0], [15.0, -5.0, -4.0], [15.0, 5.0, -4.0],
               [15.0, 15.0, -4.0], [15.0, 25.0, -8.0], [25.0, -25.0, -10.0], [25.0, -15.0, -5.0],
               [25.0, -5.0, 2.0], [25.0, 5.0, 2.0], [25.0, 15.0, -5.0], [25.0, 25.0, -10.0]]

    # Set control points
    surf.set_ctrlpts(ctrlpts, 6, 6)

    # Set knot vectors
    surf.knotvector_u = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]

    return surf


def test_bspline_curve_name(spline_surf):
    spline_surf.name = "Surface Testing"
    assert spline_surf.name == "Surface Testing"


def test_bspline_surface_degree_u(spline_surf):
    assert spline_surf.degree_u == 3


def test_bspline_surface_degree_v(spline_surf):
    assert spline_surf.degree_v == 3


def test_bspline_surface_ctrlpts(spline_surf):
    assert spline_surf.ctrlpts2d[1][1] == [-15.0, -15.0, -4.0]
    assert spline_surf.dimension == 3


def test_bspline_surface_knot_vector_u(spline_surf):
    assert spline_surf.knotvector_u == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


def test_bspline_surface_knot_vector_v(spline_surf):
    assert spline_surf.knotvector_v == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


@mark.parametrize("param, res", [
    ((0.0, 0.0), (-25.0, -25.0, -10.0)),
    ((0.0, 0.2), (-25.0, -11.403, -3.385)),
    ((0.0, 1.0), (-25.0, 25.0, -10.0)),
    ((0.3, 0.0), (-7.006, -25.0, -5.725)),
    ((0.3, 0.4), [-7.006, -3.308, -6.265]),
    ((0.3, 1.0), [-7.006, 25.0, -5.725]),
    ((0.6, 0.0), (3.533, -25.0, -4.224)),
    ((0.6, 0.6), (3.533, 3.533, -6.801)),
    ((0.6, 1.0), (3.533, 25.0, -4.224)),
    ((1.0, 0.0), (25.0, -25.0, -10.0)),
    ((1.0, 0.8), (25.0, 11.636, -2.751)),
    ((1.0, 1.0), (25.0, 25.0, -10.0))
])
def test_bspline_surface_eval(spline_surf, param, res):
    evalpt = spline_surf.evaluate_single(param)
    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


def test_bspline_surface_deriv(spline_surf):
    der1 = spline_surf.derivatives(u=0.35, v=0.35, order=2)
    spline_surf.evaluator = evaluators.SurfaceEvaluator2()
    der2 = spline_surf.derivatives(u=0.35, v=0.35, order=2)
    for k in range(0, 3):
        for l in range(0, 3 - k):
            assert abs(der1[k][l][0] - der2[k][l][0]) < GEOMDL_DELTA
            assert abs(der1[k][l][1] - der2[k][l][1]) < GEOMDL_DELTA
            assert abs(der1[k][l][2] - der2[k][l][2]) < GEOMDL_DELTA


@mark.parametrize("params, uv, res", [
    (dict(u=0.3, v=0.4), (0.3, 0.4), (-7.006, -3.308, -6.265)),
    (dict(u=0.3, num_u=2), (0.3, 0.4), (-7.006, -3.308, -6.265)),
    (dict(v=0.3, num_v=2), (0.3, 0.4), (-7.006, -3.308, -6.265))
])
def test_bspline_surface_insert_knot_eval(spline_surf, params, uv, res):
    # Insert knot
    spline_surf.insert_knot(**params)

    # Evaluate surface
    evalpt = spline_surf.evaluate_single(uv)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


@mark.parametrize("params, idx, val", [
    (dict(v=0.3, num_v=2), 4, 0.3),
    (dict(v=0.3, num_v=2), 6, 0.33)
])
def test_bspline_surface_insert_knot_kv_v(spline_surf, params, idx, val):
    # Insert knot
    spline_surf.insert_knot(**params)

    assert spline_surf.knotvector_v[idx] == val


@mark.parametrize("params, idx, val", [
    (dict(u=0.33, num_u=2), 3, 0.0),
    (dict(u=0.33, num_u=1), 6, 0.66)
])
def test_bspline_surface_insert_kv_u(spline_surf, params, idx, val):
    # Insert knot
    spline_surf.insert_knot(**params)

    assert spline_surf.knotvector_u[idx] == val


@mark.parametrize("param, num_remove", [
    (0.33, 1),
    (0.66, 1)
])
def test_bspline_surface_remove_knot_u(spline_surf, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_surf.knotvector_u)
    c_pre = spline_surf.ctrlpts_size_u
    spline_surf.remove_knot(u=param, num_u=num_remove)
    s_post = helpers.find_multiplicity(param, spline_surf.knotvector_u)
    c_post = spline_surf.ctrlpts_size_u

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


@mark.parametrize("param, num_remove", [
    (0.33, 1),
    (0.66, 1)
])
def test_bspline_surface_remove_knot_v(spline_surf, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_surf.knotvector_v)
    c_pre = spline_surf.ctrlpts_size_v
    spline_surf.remove_knot(v=param, num_v=num_remove)
    s_post = helpers.find_multiplicity(param, spline_surf.knotvector_v)
    c_post = spline_surf.ctrlpts_size_v

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


def test_bspline_surface_remove_knot_kv_u(spline_surf):
    spline_surf.remove_knot(u=0.66, num_u=1)
    s = helpers.find_multiplicity(0.66, spline_surf.knotvector_u)

    assert 0.66 not in spline_surf.knotvector_u
    assert s == 0


def test_bspline_surface_remove_knot_kv_v(spline_surf):
    spline_surf.remove_knot(v=0.33, num_v=1)
    s = helpers.find_multiplicity(0.33, spline_surf.knotvector_v)

    assert 0.33 not in spline_surf.knotvector_v
    assert s == 0


@fixture
def nurbs_surf(spline_surf):
    surf = convert.bspline_to_nurbs(spline_surf)
    return surf


def test_nurbs_weights(nurbs_surf):
    assert len(nurbs_surf.weights) == nurbs_surf.ctrlpts_size
    assert nurbs_surf.weights[5] == 1.0


@mark.parametrize("param, res", [
    ((0.0, 0.0), (-25.0, -25.0, -10.0)),
    ((0.0, 0.2), (-25.0, -11.403, -3.385)),
    ((0.0, 1.0), (-25.0, 25.0, -10.0)),
    ((0.3, 0.0), (-7.006, -25.0, -5.725)),
    ((0.3, 0.4), [-7.006, -3.308, -6.265]),
    ((0.3, 1.0), [-7.006, 25.0, -5.725]),
    ((0.6, 0.0), (3.533, -25.0, -4.224)),
    ((0.6, 0.6), (3.533, 3.533, -6.801)),
    ((0.6, 1.0), (3.533, 25.0, -4.224)),
    ((1.0, 0.0), (25.0, -25.0, -10.0)),
    ((1.0, 0.8), (25.0, 11.636, -2.751)),
    ((1.0, 1.0), (25.0, 25.0, -10.0))
])
def test_nurbs_surface_eval(nurbs_surf, param, res):
    evalpt = nurbs_surf.evaluate_single(param)
    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


@mark.parametrize("param, order, res", [
    ((0.0, 0.25), 1, [[[-25.0, -9.0771, -2.3972], [5.5511e-15, 43.6910, 17.5411]], [[90.9090, 0.0, -15.0882],[-5.9750e-15, 0.0, -140.0367]]]),
    ((0.95, 0.75), 2, [[[20.8948, 9.3097, -2.4845], [-1.1347e-14, 43.7672, -15.0153], [-5.0393e-30, 100.1022, -74.1165]], [[76.2308, -1.6965e-15, 18.0372], [9.8212e-15, -5.9448e-15, -158.5462], [4.3615e-30, -2.4356e-13, -284.3037]], [[224.5342, -5.6794e-14, 93.3843], [4.9856e-14, -4.0400e-13, -542.6274], [2.2140e-29, -1.88662e-12, -318.8808]]])
])
def test_nurbs_surface_deriv(nurbs_surf, param, order, res):
    deriv = nurbs_surf.derivatives(*param, order=order)

    for computed, expected in zip(deriv, res):
        for idx in range(order + 1):
            for c, e in zip(computed[idx], expected[idx]):
                assert abs(c - e) < GEOMDL_DELTA


def test_surface_bounding_box(spline_surf):
    # Evaluate bounding box
    to_check = spline_surf.bbox

    # Evaluation result
    result = ((-25.0, -25.0, -10.0), (25.0, 25.0, 2.0))

    assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
    assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
    assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA
    assert abs(to_check[1][0] - result[1][0]) < GEOMDL_DELTA
    assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
    assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA



"""
.. module:: VisPlotly
    :platform: Unix, Windows
    :synopsis: Plotly visualization component for NURBS-Python

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

from . import vis
import numpy as np
from plotly import graph_objs


class VisConfig(vis.VisConfigAbstract):
    """ Configuration class for Plotly visualization module.

    This class is only required when you would like to change the visual defaults of the plots and the figure,
    such as hiding control points plot or legend.

    The ``VisPlotly`` module has the following configuration variables:

    * ``ctrlpts`` (bool): Control points polygon/grid visibility. *Default: True*
    * ``evalpts`` (bool): Curve/surface points visibility. *Default: True*
    * ``bbox`` (bool): Bounding box visibility. *Default: False*
    * ``legend`` (bool): Figure legend visibility. *Default: True*
    * ``axes`` (bool): Axes and figure grid visibility. *Default: True*
    * ``trims`` (bool): Trim curves visibility. *Default: True*
    * ``axes_equal`` (bool): Enables or disables equal aspect ratio for the axes. *Default: True*
    * ``line_width`` (int): Thickness of the lines on the figure. *Default: 2*
    * ``figure_size`` (list): Size of the figure in (x, y). *Default: [800, 600]*
    * ``trim_size`` (int): Size of the trim curves. *Default: 20*

    The following example illustrates the usage of the configuration class.

    .. code-block:: python
        :linenos:

        # Create a surface (or a curve) instance
        surf = NURBS.Surface()

        # Skipping degree, knot vector and control points assignments

        # Create a visualization configuration instance with no legend, no axes and no control points grid
        vis_config = VisPlotly.VisConfig(legend=False, axes=False, ctrlpts=False)

        # Create a visualization method instance using the configuration above
        vis_obj = VisPlotly.VisSurface(vis_config)

        # Set the visualization method of the surface object
        surf.vis = vis_obj

        # Plot the surface
        surf.render()

    Please refer to the **Examples Repository** for more details.
    """
    def __init__(self, **kwargs):
        super(VisConfig, self).__init__(**kwargs)
        self.dtype = np.float
        # Set Plotly custom variables
        self.figure_image_filename = "temp-plot.html"

        # Get keyword arguments
        self.display_ctrlpts = kwargs.get('ctrlpts', True)
        self.display_evalpts = kwargs.get('evalpts', True)
        self.display_bbox = kwargs.get('bbox', False)
        self.display_trims = kwargs.get('trims', True)
        self.display_legend = kwargs.get('legend', True)
        self.display_axes = kwargs.get('axes', True)
        self.axes_equal = kwargs.get('axes_equal', True)
        self.figure_size = kwargs.get('figure_size', [1024, 768])
        self.trim_size = kwargs.get('trim_size', 1)
        self.line_width = kwargs.get('line_width', 2)


class VisCurve2D(vis.VisAbstract):
    """ Plotly visualization module for 2D curves. """
    def __init__(self, config=VisConfig(), **kwargs):
        super(VisCurve2D, self).__init__(config, **kwargs)

    def render(self, **kwargs):
        """ Plots the curve and the control points polygon. """
        # Calling parent function
        super(VisCurve2D, self).render(**kwargs)

        # Initialize variables
        plot_data = []

        for plot in self._plots:
            pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)

            # Plot control points
            if plot['type'] == 'ctrlpts' and self.vconf.display_ctrlpts:
                figure = graph_objs.Scatter(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    name=plot['name'],
                    mode='lines+markers',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dash'
                    )
                )
                plot_data.append(figure)

            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                figure = graph_objs.Scatter(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    name=plot['name'],
                    mode='lines',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width
                    )
                )
                plot_data.append(figure)

            # Plot bounding box
            if plot['type'] == 'bbox' and self.vconf.display_bbox:
                figure = graph_objs.Scatter(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    name=plot['name'],
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dashdot',
                    )
                )
                plot_data.append(figure)

            # Plot extras
            if plot['type'] == 'extras':
                figure = graph_objs.Scatter(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'][0],
                        size=plot['color'][1],
                        line=dict(
                            width=self.vconf.line_width
                        )
                    )
                )
                plot_data.append(figure)

        plot_layout = dict(
            width=self.vconf.figure_size[0],
            height=self.vconf.figure_size[1],
            autosize=False,
            showlegend=self.vconf.display_legend,
            yaxis=dict(
                scaleanchor="x",
                showgrid=self.vconf.display_axes,
                showline=self.vconf.display_axes,
                zeroline=self.vconf.display_axes,
                showticklabels=self.vconf.display_axes,
            ),
            xaxis=dict(
                showgrid=self.vconf.display_axes,
                showline=self.vconf.display_axes,
                zeroline=self.vconf.display_axes,
                showticklabels=self.vconf.display_axes,
            )
        )

        # Generate the figure
        fig = graph_objs.Figure(data=plot_data, layout=plot_layout)

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            fig.write_html(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)
        else:
            fig.write_image(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)


class VisCurve3D(vis.VisAbstract):
    """ Plotly visualization module for 3D curves. """
    def __init__(self, config=VisConfig(), **kwargs):
        super(VisCurve3D, self).__init__(config, **kwargs)

    def render(self, **kwargs):
        """ Plots the curve and the control points polygon. """
        # Calling parent function
        super(VisCurve3D, self).render(**kwargs)

        # Initialize variables
        plot_data = []

        for plot in self._plots:
            pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)

            # Try not to fail if the input is 2D
            if pts.shape[1] == 2:
                pts = np.c_[pts, np.zeros(pts.shape[0])]

            # Plot control points
            if plot['type'] == 'ctrlpts' and self.vconf.display_ctrlpts:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='lines+markers',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dash'
                    ),
                    marker=dict(
                        color=plot['color'],
                        size=self.vconf.line_width * 2,
                    )
                )
                plot_data.append(figure)

            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='lines',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width
                    ),
                )
                plot_data.append(figure)

            # Plot bounding box
            if plot['type'] == 'bbox' and self.vconf.display_bbox:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='lines',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dashdot',
                    ),
                )
                plot_data.append(figure)

            # Plot extras
            if plot['type'] == 'extras':
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'][0],
                        size=plot['color'][1],
                        line=dict(
                            width=self.vconf.line_width
                        )
                    )
                )
                plot_data.append(figure)

        plot_layout = dict(
            width=self.vconf.figure_size[0],
            height=self.vconf.figure_size[1],
            autosize=False,
            showlegend=self.vconf.display_legend,
            scene=dict(
                xaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                yaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                zaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
            ),
        )

        # Set aspect ratio
        if self.vconf.axes_equal:
            plot_layout['scene']['aspectmode'] = 'data'

        # Generate the figure
        fig = graph_objs.Figure(data=plot_data, layout=plot_layout)

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            fig.write_html(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)
        else:
            fig.write_image(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)


class VisSurface(vis.VisAbstract):
    """ Plotly visualization module for surfaces.

    Triangular mesh plot for the surface and wireframe plot for the control points grid.
    """
    def __init__(self, config=VisConfig(), **kwargs):
        super(VisSurface, self).__init__(config, **kwargs)
        self._module_config['ctrlpts'] = "points"
        self._module_config['evalpts'] = "triangles"

    def render(self, **kwargs):
        """ Plots the surface and the control points grid. """
        # Calling parent function
        super(VisSurface, self).render(**kwargs)

        # Initialize variables
        plot_data = []

        for plot in self._plots:
            # Plot control points
            if plot['type'] == 'ctrlpts' and self.vconf.display_ctrlpts:
                pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                pts[:, 2] += self._ctrlpts_offset
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'],
                        size=self.vconf.line_width * 2,
                    )
                )
                plot_data.append(figure)

            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                vertices = plot['ptsarr'][0]
                triangles = plot['ptsarr'][1]
                pts = [v.data for v in vertices]
                tri = [t.data for t in triangles]
                pts = np.array(pts, dtype=self.vconf.dtype)
                tri = np.array(tri, dtype=self.vconf.dtype)
                figure = graph_objs.Mesh3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    i=tri[:, 0],
                    j=tri[:, 1],
                    k=tri[:, 2],
                    color=plot['color'],
                    opacity=0.75,
                )
                plot_data.append(figure)

            # Plot bounding box
            if plot['type'] == 'bbox' and self.vconf.display_bbox:
                pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='lines',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dashdot',
                    ),
                )
                plot_data.append(figure)

            # Plot trim curves
            if self.vconf.display_trims:
                if plot['type'] == 'trimcurve':
                    pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                    figure = graph_objs.Scatter3d(
                        x=pts[:, 0],
                        y=pts[:, 1],
                        z=pts[:, 2],
                        name=plot['name'],
                        mode='markers',
                        marker=dict(
                            color=plot['color'],
                            size=self.vconf.trim_size * 2,
                        ),
                    )
                    plot_data.append(figure)

            # Plot extras
            if plot['type'] == 'extras':
                pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'][0],
                        size=plot['color'][1],
                        line=dict(
                            width=self.vconf.line_width
                        )
                    )
                )
                plot_data.append(figure)

        plot_layout = dict(
            width=self.vconf.figure_size[0],
            height=self.vconf.figure_size[1],
            autosize=False,
            showlegend=self.vconf.display_legend,
            scene=dict(
                xaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                yaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                zaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
            ),
        )

        # Set aspect ratio
        if self.vconf.axes_equal:
            plot_layout['scene']['aspectmode'] = 'data'

        # Generate the figure
        fig = graph_objs.Figure(data=plot_data, layout=plot_layout)

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            fig.write_html(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)
        else:
            fig.write_image(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)


class VisVolume(vis.VisAbstract):
    """ Plotly visualization module for volumes. """
    def __init__(self, config=VisConfig(), **kwargs):
        super(VisVolume, self).__init__(config, **kwargs)
        self._module_config['ctrlpts'] = "points"
        self._module_config['evalpts'] = "points"

    def render(self, **kwargs):
        """ Plots the evaluated and the control points. """
        # Calling parent function
        super(VisVolume, self).render(**kwargs)

        # Initialize variables
        plot_data = []

        for plot in self._plots:
            pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
            # Plot control points
            if plot['type'] == 'ctrlpts' and self.vconf.display_ctrlpts:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'],
                        size=self.vconf.line_width,
                    )
                )
                plot_data.append(figure)

            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'],
                        size=self.vconf.line_width * 2,
                    )
                )
                plot_data.append(figure)

            # Plot bounding box
            if plot['type'] == 'bbox' and self.vconf.display_bbox:
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='lines',
                    line=dict(
                        color=plot['color'],
                        width=self.vconf.line_width,
                        dash='dashdot',
                    ),
                )
                plot_data.append(figure)

            # Plot extras
            if plot['type'] == 'extras':
                figure = graph_objs.Scatter3d(
                    x=pts[:, 0],
                    y=pts[:, 1],
                    z=pts[:, 2],
                    name=plot['name'],
                    mode='markers',
                    marker=dict(
                        color=plot['color'][0],
                        size=plot['color'][1],
                        line= dict(
                            width=self.vconf.line_width
                        )
                    )
                )
                plot_data.append(figure)

        plot_layout = dict(
            width=self.vconf.figure_size[0],
            height=self.vconf.figure_size[1],
            autosize=False,
            showlegend=self.vconf.display_legend,
            scene=dict(
                xaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                yaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
                zaxis=dict(
                    showgrid=self.vconf.display_axes,
                    showline=self.vconf.display_axes,
                    zeroline=self.vconf.display_axes,
                    showticklabels=self.vconf.display_axes,
                    title='',
                ),
            ),
        )

        # Set aspect ratio
        if self.vconf.axes_equal:
            plot_layout['scene']['aspectmode'] = 'data'

        # Generate the figure
        fig = graph_objs.Figure(data=plot_data, layout=plot_layout)

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            fig.write_html(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)
        else:
            fig.write_image(file=self.vconf.figure_image_filename if fig_filename is None else fig_filename)



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018 Onur Rauf Bingol

    Tests geomdl.utilities module. Requires "pytest" to run.
"""

import pytest
from geomdl import utilities
from geomdl import knotvector
from geomdl import control_points
from geomdl import utilities
from geomdl.exceptions import GeomdlException

GEOMDL_DELTA = 10e-6


def test_generate_knot_vector1():
    with pytest.raises(ValueError):
        degree = 0
        num_ctrlpts = 12
        knotvector.generate(degree, num_ctrlpts)


def test_generate_knot_vector2():
    with pytest.raises(ValueError):
        degree = 4
        num_ctrlpts = 0
        knotvector.generate(degree, num_ctrlpts)


def test_generate_knot_vector3():
    with pytest.raises(ValueError):
        degree = 0
        num_ctrlpts = 0
        knotvector.generate(degree, num_ctrlpts)


def test_generate_knot_vector4():
    degree = 4
    num_ctrlpts = 12
    autogen_kv = knotvector.generate(degree, num_ctrlpts)
    result = [0.0, 0.0, 0.0, 0.0, 0.0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1.0, 1.0, 1.0, 1.0, 1.0]
    assert autogen_kv == result


def test_generate_knot_vector5():
    # testing auto-generated unclamped knot vector
    degree = 3
    num_ctrlpts = 5
    result = [0.0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1.0]
    autogen_kv = knotvector.generate(degree, num_ctrlpts, clamped=False)
    assert autogen_kv == result


def test_check_knot_vector1():
    with pytest.raises(ValueError):
        knotvector.check(4, tuple(), 12)


def test_check_knot_vector2():
    to_check = knotvector.check(4, (1, 2, 3, 4), 12)
    result = False
    assert to_check == result


def test_check_knot_vector3():
    to_check = knotvector.check(3, (5, 3, 6, 5, 4, 5, 6), 3)
    result = False
    assert to_check == result


def test_check_knot_vector4():
    degree = 4
    num_ctrlpts = 12
    autogen_kv = knotvector.generate(degree, num_ctrlpts)
    check_result = knotvector.check(degree=degree, num_ctrlpts=num_ctrlpts, knot_vector=autogen_kv)
    assert check_result


def test_check_knot_vector5():
    degree = 4
    num_ctrlpts = 12
    with pytest.raises(TypeError):
        knotvector.check(degree=degree, num_ctrlpts=num_ctrlpts, knot_vector=5)


def test_normalize_knot_vector1():
    # check for empty list/tuple
    with pytest.raises(ValueError):
        knotvector.normalize(tuple())


def test_normalize_knot_vector2():
    input_kv = (-5, -5, -3, -2, 2, 3, 5, 5)
    output_kv = [0.0, 0.0, 0.2, 0.3, 0.7, 0.8, 1.0, 1.0]
    to_check = knotvector.normalize(input_kv)
    assert to_check == output_kv


def test_normalize_knot_vector3():
    with pytest.raises(TypeError):
        knotvector.normalize(5)


def test_check_uv1():
    u = -0.1
    v = 0.1
    assert not utilities.check_params([u, v])


def test_check_uv2():
    u = 2
    v = 0.1
    assert not utilities.check_params([u, v])


def test_check_uv3():
    v = -0.1
    u = 0.1
    assert not utilities.check_params([u, v])


def test_check_uv4():
    v = 2
    u = 0.1
    assert not utilities.check_params([u, v])


def test_color_generator():
    seed = 17  # some number to be used as the random seed
    result = utilities.color_generator(seed)
    to_check = utilities.color_generator(seed)
    assert to_check == result


def test_cpman_curve1():
    """ Control Points Manager: get-set point (curve) """
    pt = [0.0, 0.2, 0.3]
    p = 3
    sz = 10
    cpman = control_points.CurveManager(sz)
    cpman.set_ctrlpt(pt, p)
    assert cpman.get_ctrlpt(3) == pt


def test_cpman_curve2():
    """ Control Points Manager: get empty point (curve) """
    p = 5
    sz = 12
    cpman = control_points.CurveManager(sz)
    assert cpman.get_ctrlpt(p) == list()


def test_cpman_curve3():
    """ Control Points Manager: check for invalid index """
    p = 12
    sz = 5
    cpman = control_points.CurveManager(sz)
    assert cpman.get_ctrlpt(p) == None


def test_cpman_curve4():
    """ Control Points Manager: get-set attachment (valid, list) """
    d = [0.0, 1.0, 2.0, 3.0]
    p = 5
    sz = 12
    cpman = control_points.CurveManager(sz, testdata=4)
    cpman.set_ptdata(dict(testdata=d), p)
    retv1 = cpman.get_ptdata('testdata', p)
    retv2 = cpman.get_ptdata('testdata', p + 1)
    assert retv1[2] == 2.0
    assert retv2[2] == 0.0


def test_cpman_curve5():
    """ Control Points Manager: get-set attachment (invalid, list) """
    d = [0.0, 1.0, 2.0, 3.0]
    p = 5
    sz = 12
    cpman = control_points.CurveManager(sz, testdata=4)
    cpman.set_ptdata(dict(testdata=d), p)
    retv = cpman.get_ptdata('testdata2', p)
    assert retv == None


def test_cpman_curve6():
    """ Control Points Manager: get-set attachment (exception) """
    with pytest.raises(GeomdlException):
        d = [0.0, 1.0, 2.0, 3.0]
        p = 5
        sz = 12
        cpman = control_points.CurveManager(sz, testdata=3)
        cpman.set_ptdata(dict(testdata=d), p)


def test_cpman_curve7():
    """ Control Points Manager: get-set attachment (valid, float) """
    d = 13
    p = 5
    sz = 12
    cpman = control_points.CurveManager(sz, testdata=1)
    cpman.set_ptdata(dict(testdata=d), p)
    assert cpman.get_ptdata('testdata', 5) == 13


def test_cpman_curve8():
    """ Control Points Manager: try to set invalid key """
    with pytest.raises(GeomdlException):
        d = [0.0, 1.0, 2.0, 3.0]
        p = 5
        sz = 12
        cpman = control_points.CurveManager(sz, testdata=4)
        cpman.set_ptdata({'testdata1': d}, p)


def test_cpman_surface1():
    """ Control Points Manager: get-set point (surface) """
    pt = [1.0, 2.0, 3.0]
    p = [2 ,3]
    sz = [4, 3]
    cpman = control_points.SurfaceManager(*sz)
    cpman.set_ctrlpt(pt, *p)
    assert cpman.get_ctrlpt(2, 3) == pt


def test_cpman_volume1():
    """ Control Points Manager: get-set point (volume) """
    pt = [1.0, 2.0, 3.0]
    p = [2, 3, 1]
    sz = [4, 3, 2]
    cpman = control_points.VolumeManager(*sz)
    cpman.set_ctrlpt(pt, *p)
    assert cpman.get_ctrlpt(2, 3, 1) == pt



"""
.. module:: _shortcuts
    :platform: Unix, Windows
    :synopsis: Provides shortcut functions for creating new instances of the geomdl classes

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

from . import BSpline
from . import NURBS
from . import multi
from . import freeform


# Initialize an empty __all__ for controlling imports
__all__ = []


def generate_curve(rational=False):
    if rational:
        return NURBS.Curve()
    return BSpline.Curve()


def generate_surface(rational=False):
    if rational:
        return NURBS.Surface()
    return BSpline.Surface()


def generate_volume(rational=False):
    if rational:
        return NURBS.Volume()
    return BSpline.Volume()


def generate_freeform():
    return freeform.Freeform()


def generate_container_curve():
    return multi.CurveContainer()


def generate_container_surface():
    return multi.SurfaceContainer()


def generate_container_volume():
    return multi.VolumeContainer()



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018 Onur Rauf Bingol

    Tests visualization modules. Requires "pytest" to run.
"""

import os
import pytest
from geomdl import BSpline
from geomdl import multi
from geomdl import operations

import matplotlib
matplotlib.use('agg')
from geomdl.visualization import VisMPL

SAMPLE_SIZE = 5


@pytest.fixture
def bspline_curve2d():
    """ Creates a 2-dimensional B-Spline curve instance """
    # Create a curve instance
    curve = BSpline.Curve()

    # Set curve degree
    curve.degree = 3

    # Set control points
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]

    # Set knot vector
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]

    # Set sample size
    curve.sample_size = SAMPLE_SIZE

    # Return the instance
    return curve


@pytest.fixture
def bspline_curve3d():
    """ Creates a 3-dimensional B-Spline curve instance """
    # Create a curve instance
    curve = BSpline.Curve()

    # Set curve degree
    curve.degree = 4

    # Set control points
    curve.ctrlpts = [[5.0, 15.0, 0.0], [10.0, 25.0, 5.0], [20.0, 20.0, 10.0], [15.0, -5.0, 15.0], [7.5, 10.0, 20.0],
                     [12.5, 15.0, 25.0], [15.0, 0.0, 30.0], [5.0, -10.0, 35.0], [10.0, 15.0, 40.0], [5.0, 15.0, 30.0]]

    # Set knot vector
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0, 1.0, 1.0, 1.0, 1.0]

    # Set sample size
    curve.sample_size = SAMPLE_SIZE

    # Return the instance
    return curve


@pytest.fixture
def bspline_surface():
    """ Creates a B-Spline surface instance """
    # Create a surface instance
    surf = BSpline.Surface()

    # Set degrees
    surf.degree_u = 3
    surf.degree_v = 3

    # Set control points
    surf.ctrlpts_size_u = 6
    surf.ctrlpts_size_v = 6
    surf.ctrlpts = [[-25.0, -25.0, -10.0], [-25.0, -15.0, -5.0], [-25.0, -5.0, 0.0], [-25.0, 5.0, 0.0],
                    [-25.0, 15.0, -5.0], [-25.0, 25.0, -10.0], [-15.0, -25.0, -8.0], [-15.0, -15.0, -4.0],
                    [-15.0, -5.0, -4.0], [-15.0, 5.0, -4.0], [-15.0, 15.0, -4.0], [-15.0, 25.0, -8.0],
                    [-5.0, -25.0, -5.0], [-5.0, -15.0, -3.0], [-5.0, -5.0, -8.0], [-5.0, 5.0, -8.0],
                    [-5.0, 15.0, -3.0], [-5.0, 25.0, -5.0], [5.0, -25.0, -3.0], [5.0, -15.0, -2.0],
                    [5.0, -5.0, -8.0], [5.0, 5.0, -8.0], [5.0, 15.0, -2.0], [5.0, 25.0, -3.0],
                    [15.0, -25.0, -8.0], [15.0, -15.0, -4.0], [15.0, -5.0, -4.0], [15.0, 5.0, -4.0],
                    [15.0, 15.0, -4.0], [15.0, 25.0, -8.0], [25.0, -25.0, -10.0], [25.0, -15.0, -5.0],
                    [25.0, -5.0, 2.0], [25.0, 5.0, 2.0], [25.0, 15.0, -5.0], [25.0, 25.0, -10.0]]

    # Set knot vectors
    surf.knotvector_u = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]

    # Set sample size
    surf.sample_size = SAMPLE_SIZE

    # Return the instance
    return surf


# Test if plotting a 2-dimensional curve without a window is possible
def test_curve2d_fig_nowindow(bspline_curve2d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve3D(config=conf)

    fname = conf.figure_image_filename

    bspline_curve2d.vis = vis
    bspline_curve2d.render(plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_curve2d_fig_save(bspline_curve2d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve2D(config=conf)

    fname = "test-curve.png"

    bspline_curve2d.vis = vis
    bspline_curve2d.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test if plotting a 2-dimensional multi-curve without a window is possible
def test_curve2d_multi_fig_nowindow(bspline_curve2d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve2D(config=conf)

    fname = conf.figure_image_filename

    data = operations.decompose_curve(bspline_curve2d)
    multi_shape = multi.CurveContainer(data)
    multi_shape.vis = vis
    multi_shape.render(plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_curve2d_multi_fig_save(bspline_curve2d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve2D(config=conf)

    fname = "test-multi_curve.png"

    data = operations.decompose_curve(bspline_curve2d)
    multi_shape = multi.CurveContainer(data)
    multi_shape.vis = vis
    multi_shape.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test if plotting a 3-dimensional curve without a window is possible
def test_curve3d_fig_nowindow(bspline_curve3d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve2D(config=conf)

    fname = conf.figure_image_filename

    bspline_curve3d.vis = vis
    bspline_curve3d.render(plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_curve3d_fig_save(bspline_curve3d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve3D(config=conf)

    fname = "test-curve.png"

    bspline_curve3d.vis = vis
    bspline_curve3d.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test if plotting a 3-dimensional multi-curve without a window is possible
def test_curve3d_multi_fig_nowindow(bspline_curve3d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve3D(config=conf)

    data = operations.decompose_curve(bspline_curve3d)
    multi_shape = multi.CurveContainer(data)
    multi_shape.vis = vis
    multi_shape.render(plot=False)

    assert os.path.isfile(conf.figure_image_filename)
    assert os.path.getsize(conf.figure_image_filename) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_curve3d_multi_fig_save(bspline_curve3d):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisCurve3D(config=conf)

    fname = "test-multi_curve.png"

    data = operations.decompose_curve(bspline_curve3d)
    multi_shape = multi.CurveContainer(data)
    multi_shape.vis = vis
    multi_shape.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test if plotting a surface without a window is possible
def test_surf_fig_nowindow(bspline_surface):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisSurface(config=conf)

    fname = conf.figure_image_filename

    bspline_surface.vis = vis
    bspline_surface.render(plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_surf_fig_save(bspline_surface):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisSurface(config=conf)

    fname = "test-surface.png"

    bspline_surface.vis = vis
    bspline_surface.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test offsetting control points grid
def test_surf_ctrlpts_offset(bspline_surface):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisSurface(config=conf)

    # Set control points grid offset
    vis.ctrlpts_offset = 3.5

    fname = "test-surface.png"

    bspline_surface.vis = vis
    bspline_surface.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Test if plotting a multi-surface without a window is possible
def test_surf_multi_fig_nowindow(bspline_surface):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisSurfScatter(config=conf)

    fname = conf.figure_image_filename

    data = operations.decompose_surface(bspline_surface)
    multi_shape = multi.SurfaceContainer(data)
    multi_shape.vis = vis
    multi_shape.render(plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(conf.figure_image_filename):
        os.remove(conf.figure_image_filename)


# Test if using a different file name is possible
def test_surf_multi_fig_save(bspline_surface):
    conf = VisMPL.VisConfig()
    vis = VisMPL.VisSurfWireframe(config=conf)

    fname = "test-multi_surface.png"

    data = operations.decompose_surface(bspline_surface)
    multi_shape = multi.SurfaceContainer(data)
    multi_shape.vis = vis
    multi_shape.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_deriv_curve_fig(bspline_curve2d):
    fname = "test-derivative_curve.png"

    data = operations.derivative_curve(bspline_curve2d)
    multi_shape = multi.CurveContainer(data)
    multi_shape.vis = VisMPL.VisCurve2D()
    multi_shape.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_deriv_surf_fig(bspline_surface):
    fname = "test-derivative_surface.png"

    data = operations.derivative_surface(bspline_surface)
    multi_shape = multi.SurfaceContainer(data)
    multi_shape.vis = VisMPL.VisSurface()
    multi_shape.render(filename=fname, plot=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018 Onur Rauf Bingol

    Tests file I/O operations. Requires "pytest" to run.
"""

import os
import pytest

from geomdl import BSpline, NURBS
from geomdl import multi
from geomdl import exchange
from geomdl import exchange_vtk
from geomdl import compatibility
from geomdl import operations

FILE_NAME = 'testing'
SAMPLE_SIZE = 25


@pytest.fixture
def bspline_curve3d():
    """ Creates a B-Spline 3-dimensional curve instance """
    curve = BSpline.Curve()
    curve.degree = 2
    curve.ctrlpts = [[1, 1, 0], [2, 1, -1], [2, 2, 0]]
    curve.knotvector = [0, 0, 0, 1, 1, 1]
    return curve


@pytest.fixture
def bspline_surface():
    """ Creates a B-Spline surface instance """
    surf = BSpline.Surface()
    surf.degree_u = 2
    surf.degree_v = 2
    surf.ctrlpts_size_u = 3
    surf.ctrlpts_size_v = 3
    surf.ctrlpts = [[0, 0, 0], [0, 1, 0], [0, 2, -3],
                    [1, 0, 6], [1, 1, 0], [1, 2, 0],
                    [2, 0, 0], [2, 1, 0], [2, 2, 3]]
    surf.knotvector_u = [0, 0, 0, 1, 1, 1]
    surf.knotvector_v = [0, 0, 0, 1, 1, 1]
    return surf


@pytest.fixture
def nurbs_surface():
    """ Creates a NURBS surface instance """
    surf = NURBS.Surface()
    surf.degree_u = 2
    surf.degree_v = 2
    surf.ctrlpts_size_u = 3
    surf.ctrlpts_size_v = 3
    surf.ctrlpts = [[0, 0, 0], [0, 1, 0], [0, 2, -3],
                    [1, 0, 6], [1, 1, 0], [1, 2, 0],
                    [2, 0, 0], [2, 1, 0], [2, 2, 3]]
    # use the auto-generated weights vector
    surf.knotvector_u = [0, 0, 0, 1, 1, 1]
    surf.knotvector_v = [0, 0, 0, 1, 1, 1]
    return surf


@pytest.fixture
def nurbs_surface_decompose():
    """ Creates a NURBS surface instance (decomposable) """
    surf = NURBS.Surface()
    surf.degree_u = 2
    surf.degree_v = 2
    surf.ctrlpts_size_u = 3
    surf.ctrlpts_size_v = 4
    surf.ctrlpts = [[0, 0, 0], [0, 1, 0], [0, 2, -3], [0, 3, 7],
                    [1, 0, 6], [1, 1, 0], [1, 2, 0], [1, 3, 8],
                    [2, 0, 0], [2, 1, 0], [2, 2, 3], [1, 3, 7]]
    # use the auto-generated weights vector
    surf.knotvector_u = [0, 0, 0, 1, 1, 1]
    surf.knotvector_v = [0, 0, 0, 0.5, 1, 1, 1]
    return surf


# Tests if the .obj file exists
def test_export_obj_single(nurbs_surface):
    fname = FILE_NAME + ".obj"

    nurbs_surface.sample_size = SAMPLE_SIZE
    exchange.export_obj(nurbs_surface, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .off file exists
def test_export_off_single(nurbs_surface):
    fname = FILE_NAME + ".off"

    nurbs_surface.sample_size = SAMPLE_SIZE
    exchange.export_off(nurbs_surface, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .stl file exists
def test_export_stl_single(nurbs_surface):
    fname = FILE_NAME + ".stl"

    nurbs_surface.sample_size = SAMPLE_SIZE
    exchange.export_stl(nurbs_surface, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .stl file exists (ascii)
def test_export_stl_ascii_single(nurbs_surface):
    fname = FILE_NAME + ".stl"

    nurbs_surface.sample_size = SAMPLE_SIZE
    exchange.export_stl(nurbs_surface, fname, binary=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .obj file exists
def test_export_obj_multi(nurbs_surface_decompose):
    fname = FILE_NAME + ".obj"

    data = operations.decompose_surface(nurbs_surface_decompose)
    nurbs_multi = multi.SurfaceContainer(data)

    nurbs_multi.sample_size = SAMPLE_SIZE
    exchange.export_obj(nurbs_multi, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .off file exists
def test_export_off_multi(nurbs_surface_decompose):
    fname = FILE_NAME + ".off"

    data = operations.decompose_surface(nurbs_surface_decompose)
    nurbs_multi = multi.SurfaceContainer(data)

    nurbs_multi.sample_size = SAMPLE_SIZE
    exchange.export_off(nurbs_multi, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .stl file exists
def test_export_stl_multi(nurbs_surface_decompose):
    fname = FILE_NAME + ".stl"

    data = operations.decompose_surface(nurbs_surface_decompose)
    nurbs_multi = multi.SurfaceContainer(data)

    nurbs_multi.sample_size = SAMPLE_SIZE
    exchange.export_stl(nurbs_multi, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Tests if the .stl file exists (ascii)
def test_export_stl_ascii_multi(nurbs_surface_decompose):
    fname = FILE_NAME + ".stl"

    data = operations.decompose_surface(nurbs_surface_decompose)
    nurbs_multi = multi.SurfaceContainer(data)

    nurbs_multi.sample_size = SAMPLE_SIZE
    exchange.export_stl(nurbs_multi, fname, binary=False)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_txt_curve(bspline_curve3d):
    fname = FILE_NAME + ".txt"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_curve3d, fname)

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_import_txt_curve(bspline_curve3d):
    fname = FILE_NAME + ".txt"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_curve3d, fname)

    # Import text file
    result = exchange.import_txt(fname)

    res_array = []
    for res in result:
        res_array.append(res)

    assert res_array == bspline_curve3d.ctrlpts

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_import_txt_surface1(bspline_surface):
    fname = FILE_NAME + ".txt"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_surface, fname, two_dimensional=False)

    # Import text file
    result = exchange.import_txt(fname, two_dimensional=False)

    res_array = []
    for res in result:
        res_array.append(res)

    assert res_array == bspline_surface.ctrlpts

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_import_txt_surface2(bspline_surface):
    fname = FILE_NAME + ".txt"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_surface, fname, two_dimensional=True)

    # Import text file
    result, size_u, size_v = exchange.import_txt(fname, two_dimensional=True)

    res_array = []
    for res in result:
        res_array.append(res)

    assert res_array == bspline_surface.ctrlpts
    assert size_u == bspline_surface.ctrlpts_size_u
    assert size_v == bspline_surface.ctrlpts_size_v

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_vtk_curve_ctrlpts(bspline_curve3d):
    fname = FILE_NAME + ".vtk"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange_vtk.export_polydata(bspline_curve3d, fname, point_type="ctrlpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_vtk_surface_ctrlpts(bspline_surface):
    fname = FILE_NAME + ".vtk"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange_vtk.export_polydata(bspline_surface, fname, point_type="ctrlpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_vtk_curve_evalpts(bspline_curve3d):
    fname = FILE_NAME + ".vtk"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange_vtk.export_polydata(bspline_curve3d, fname, point_type="evalpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_vtk_surface_evalpts(bspline_surface):
    fname = FILE_NAME + ".vtk"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange_vtk.export_polydata(bspline_surface, fname, point_type="evalpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_csv_curve_ctrlpts(bspline_curve3d):
    fname = FILE_NAME + ".csv"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange.export_csv(bspline_curve3d, fname, point_type="ctrlpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_csv_surface_ctrlpts(bspline_surface):
    fname = FILE_NAME + ".csv"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_csv(bspline_surface, fname, point_type="ctrlpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_csv_curve_evalpts(bspline_curve3d):
    fname = FILE_NAME + ".csv"

    bspline_curve3d.sample_size = SAMPLE_SIZE
    exchange.export_csv(bspline_curve3d, fname, point_type="evalpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


def test_export_csv_surface_evalpts(bspline_surface):
    fname = FILE_NAME + ".csv"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_csv(bspline_surface, fname, point_type="evalpts")

    assert os.path.isfile(fname)
    assert os.path.getsize(fname) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname):
        os.remove(fname)


# Testing read-write operations in compatibility module
def test_compatibility_flip_ctrlpts2d_file1(bspline_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = FILE_NAME + "_out.txt"

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_surface, fname_in, two_dimensional=True)

    compatibility.flip_ctrlpts2d_file(fname_in, fname_out)

    assert os.path.isfile(fname_out)
    assert os.path.getsize(fname_out) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname_in):
        os.remove(fname_in)
        os.remove(fname_out)


def test_compatibility_flip_ctrlpts2d_file2(bspline_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = FILE_NAME + "_out.txt"

    with pytest.raises(IOError):
        compatibility.flip_ctrlpts2d_file(fname_in, fname_out)


def test_compatibility_flip_ctrlpts2d_file3(bspline_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = ""

    bspline_surface.sample_size = SAMPLE_SIZE
    exchange.export_txt(bspline_surface, fname_in, two_dimensional=True)

    with pytest.raises(IOError):
        compatibility.flip_ctrlpts2d_file(fname_in, fname_out)

    # Clean up temporary file if exists
    if os.path.isfile(fname_in):
        os.remove(fname_in)


def test_compatibility_generate_ctrlpts2d_weights_file(nurbs_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = FILE_NAME + "_out.txt"

    nurbs_surface.sample_size = SAMPLE_SIZE
    exchange.export_txt(nurbs_surface, fname_in, two_dimensional=True)

    compatibility.generate_ctrlpts2d_weights_file(fname_in, fname_out)

    assert os.path.isfile(fname_out)
    assert os.path.getsize(fname_out) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname_in):
        os.remove(fname_in)
        os.remove(fname_out)


def test_compatibility_generate_ctrlptsw2d_file1(nurbs_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = FILE_NAME + "_out.txt"
    fname_final = FILE_NAME + "_final.txt"

    exchange.export_txt(nurbs_surface, fname_in, two_dimensional=True)

    compatibility.generate_ctrlpts2d_weights_file(fname_in, fname_out)
    compatibility.generate_ctrlptsw2d_file(fname_out, fname_final)

    assert os.path.isfile(fname_final)
    assert os.path.getsize(fname_final) > 0

    # Clean up temporary file if exists
    if os.path.isfile(fname_in):
        os.remove(fname_in)
        os.remove(fname_out)
        os.remove(fname_final)


def test_compatibility_generate_ctrlptsw2d_file2(nurbs_surface):
    fname_in = FILE_NAME + "_in.txt"
    fname_out = FILE_NAME + "_out.txt"
    fname_final = FILE_NAME + "_final.txt"

    exchange.export_txt(nurbs_surface, fname_in, two_dimensional=True)

    compatibility.generate_ctrlpts2d_weights_file(fname_in, fname_out)
    compatibility.generate_ctrlptsw2d_file(fname_out, fname_final)
    ctrlpts, size_u, size_v = exchange.import_txt(fname_final, two_dimensional=True)
    ctrlptsw = compatibility.generate_ctrlptsw(ctrlpts)

    res_array = []
    for res in ctrlptsw:
        res_array.append(res)

    assert nurbs_surface.ctrlptsw == res_array
    assert nurbs_surface.ctrlpts_size_u == size_u
    assert nurbs_surface.ctrlpts_size_v == size_v

    # Clean up temporary file if exists
    if os.path.isfile(fname_in):
        os.remove(fname_in)
        os.remove(fname_out)
        os.remove(fname_final)



"""
.. module:: control_points
    :platform: Unix, Windows
    :synopsis: Provides helper classes for managing control points

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

import abc
import copy
from functools import reduce
from .exceptions import GeomdlException
from ._utilities import export, add_metaclass


@add_metaclass(abc.ABCMeta)
class AbstractManager(object):
    """ Abstract base class for control points manager classes.

    Control points manager class provides an easy way to set control points without knowing
    the internal data structure of the geometry classes. The manager class is initialized
    with the number of control points in all parametric dimensions.

    All classes extending this class should implement the following methods:

    * ``find_index``

    This class provides the following properties:

    * :py:attr:`ctrlpts`

    This class provides the following methods:

    * :py:meth:`get_ctrlpt`
    * :py:meth:`set_ctrlpt`
    * :py:meth:`get_ptdata`
    * :py:meth:`set_ptdata`
    """
    __slots__ = ('_size', '_num_ctrlpts', '_attachment', '_points', '_pt_data', '_cache', '_iter_index')

    def __init__(self, *args, **kwargs):
        self._size = [int(arg) for arg in args]  # size in all parametric dimensions
        self._num_ctrlpts = reduce(lambda x, y: x *y, self._size)  # number of control points
        self._attachment = kwargs if kwargs else dict()  # data attached to the control points
        self._points = list()  # list of control points
        self._pt_data = dict()  # dict containing lists of additional data attached to the control points
        self._cache = {}
        self.reset()  # initialize control points list

    def __iter__(self):
        self._iter_index = 0
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        try:
            result = self._points[self._iter_index]
        except IndexError:
            raise StopIteration
        self._iter_index += 1
        return result

    def __len__(self):
        return len(self._points)

    def __reversed__(self):
        return reversed(self._points)

    def __getitem__(self, idx):
        return self._points[idx]

    def __setitem__(self, idx, val):
        self._points[idx] = val

    def __copy__(self):
        # Create a new instance
        cls = self.__class__
        result = cls.__new__(cls)
        # Copy all attributes
        for var in self.__slots__:
            setattr(result, var, copy.copy(getattr(self, var)))
        # Return updated instance
        return result

    def __deepcopy__(self, memo):
        # Create a new instance
        cls = self.__class__
        result = cls.__new__(cls)
        # Don't copy self reference
        memo[id(self)] = result
        # Don't copy the cache
        memo[id(self._cache)] = self._cache.__new__(dict)
        # Deep copy all other attributes
        for var in self.__slots__:
            setattr(result, var, copy.deepcopy(getattr(self, var), memo))
        # Return updated instance
        return result

    @property
    def ctrlpts(self):
        """ Control points.

        Please refer to the `wiki <https://github.com/orbingol/NURBS-Python/wiki/Using-Python-Properties>`_ for details
        on using this class member.

        :getter: Gets the control points
        :setter: Sets the control points
        """
        return self._points

    @ctrlpts.setter
    def ctrlpts(self, value):
        self._points = value

    def reset(self):
        """ Resets/initializes the internal control points array. """
        self._points[:] = [[] for _ in range(self._num_ctrlpts)]
        for k, v in self._attachment.items():
            if v > 1:
                self._pt_data[k] = [[0.0 for _ in range(v)] for _ in range(self._num_ctrlpts)]
            else:
                self._pt_data[k] = [0.0 for _ in range(self._num_ctrlpts)]

    def get_ctrlpt(self, *args):
        """ Gets the control point from the given location in the array. """
        # Find the index
        idx = self.find_index(*args)
        # Return the control point
        try:
            return self._points[idx]
        except IndexError:
            return None

    def set_ctrlpt(self, pt, *args):
        """ Puts the control point to the given location in the array.

        :param pt: control point
        :type pt: list, tuple
        """
        if not isinstance(pt, (list, tuple)):
            raise GeomdlException("'pt' argument should be a list or tuple")
        if len(args) != len(self._size):
            raise GeomdlException("Input dimensions are not compatible with the geometry")
        # Find the index
        idx = self.find_index(*args)
        # Set control point
        try:
            self._points[idx] = pt
        except IndexError:
            raise GeomdlException("Index is out of range")

    def get_ptdata(self, dkey, *args):
        """ Gets the data attached to the control point.

        :param dkey: key of the attachment dictionary
        :param dkey: str
        """
        # Find the index
        idx = self.find_index(*args)
        # Return the attached data
        try:
            return self._pt_data[dkey][idx]
        except IndexError:
            return None
        except KeyError:
            return None

    def set_ptdata(self, adct, *args):
        """ Attaches the data to the control point.

        :param adct: attachment dictionary
        :param adct: dict
        """
        # Find the index
        idx = self.find_index(*args)
        # Attach the data to the control point
        try:
            for k, val in adct.items():
                if k in self._pt_data:
                    if isinstance(val, (list, tuple)):
                        for j, v in enumerate(val):
                            self._pt_data[k][idx][j] = v
                    else:
                        self._pt_data[k][idx] = val
                else:
                    raise GeomdlException("Invalid key: " + str(k))
        except IndexError:
            raise GeomdlException("Index is out of range")

    @abc.abstractmethod
    def find_index(self, *args):
        """ Finds the array index from the given parametric positions.

        .. note::

            This is an abstract method and it must be implemented in the subclass.
        """
        return 0


class CurveManager(AbstractManager):
    """ Curve control points manager.

    Control points manager class provides an easy way to set control points without knowing
    the internal data structure of the geometry classes. The manager class is initialized
    with the number of control points in all parametric dimensions.

    B-spline curves are defined in one parametric dimension. Therefore, this manager class
    should be initialized with a single integer value.

    .. code-block:: python

        # Assuming that the curve has 10 control points
        manager = CurveManager(10)

    Getting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = spline.ctrlpts_size_u

        # Generate control points manager
        cpt_manager = control_points.SurfaceManager(size_u)
        cpt_manager.ctrlpts = spline.ctrlpts

        # Control points array to be used externally
        control_points = []

        # Get control points from the spline geometry
        for u in range(size_u):
            pt = cpt_manager.get_ctrlpt(u)
            control_points.append(pt)

    Setting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = 5

        # Create control points manager
        points = control_points.SurfaceManager(size_u)

        # Set control points
        for u in range(size_u):
            # 'pt' is the control point, e.g. [10, 15, 12]
            points.set_ctrlpt(pt, u, v)

        # Create spline geometry
        curve = BSpline.Curve()

        # Set control points
        curve.ctrlpts = points.ctrlpts
    """
    def __init__(self, *args, **kwargs):
        super(CurveManager, self).__init__(*args, **kwargs)

    def find_index(self, *args):
        super(CurveManager, self).find_index(*args)
        return args[0]


class SurfaceManager(AbstractManager):
    """ Surface control points manager.

    Control points manager class provides an easy way to set control points without knowing
    the internal data structure of the geometry classes. The manager class is initialized
    with the number of control points in all parametric dimensions.

    B-spline surfaces are defined in one parametric dimension. Therefore, this manager class
    should be initialized with two integer values.

    .. code-block:: python

        # Assuming that the surface has size_u = 5 and size_v = 7 control points
        manager = SurfaceManager(5, 7)

    Getting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = spline.ctrlpts_size_u
        size_v = spline.ctrlpts_size_v

        # Generate control points manager
        cpt_manager = control_points.SurfaceManager(size_u, size_v)
        cpt_manager.ctrlpts = spline.ctrlpts

        # Control points array to be used externally
        control_points = []

        # Get control points from the spline geometry
        for u in range(size_u):
            for v in range(size_v):
                pt = cpt_manager.get_ctrlpt(u, v)
                control_points.append(pt)

    Setting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = 5
        size_v = 3

        # Create control points manager
        points = control_points.SurfaceManager(size_u, size_v)

        # Set control points
        for u in range(size_u):
            for v in range(size_v):
                # 'pt' is the control point, e.g. [10, 15, 12]
                points.set_ctrlpt(pt, u, v)

        # Create spline geometry
        surf = BSpline.Surface()

        # Set control points
        surf.ctrlpts = points.ctrlpts
    """
    def __init__(self, *args, **kwargs):
        super(SurfaceManager, self).__init__(*args, **kwargs)

    def find_index(self, *args):
        super(SurfaceManager, self).find_index(*args)
        return args[1] + (args[0] * self._size[1])


class VolumeManager(AbstractManager):
    """ Volume control points manager.

    Control points manager class provides an easy way to set control points without knowing
    the internal data structure of the geometry classes. The manager class is initialized
    with the number of control points in all parametric dimensions.

    B-spline volumes are defined in one parametric dimension. Therefore, this manager class
    should be initialized with there integer values.

    .. code-block:: python

        # Assuming that the volume has size_u = 5, size_v = 12 and size_w = 3 control points
        manager = VolumeManager(5, 12, 3)

    Gettting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = spline.ctrlpts_size_u
        size_v = spline.ctrlpts_size_v
        size_w = spline.ctrlpts_size_w

        # Generate control points manager
        cpt_manager = control_points.SurfaceManager(size_u, size_v, size_w)
        cpt_manager.ctrlpts = spline.ctrlpts

        # Control points array to be used externally
        control_points = []

        # Get control points from the spline geometry
        for u in range(size_u):
            for v in range(size_v):
                for w in range(size_w):
                    pt = cpt_manager.get_ctrlpt(u, v, w)
                    control_points.append(pt)

    Setting the control points:

    .. code-block:: python

        # Number of control points in all parametric dimensions
        size_u = 5
        size_v = 3
        size_w = 2

        # Create control points manager
        points = control_points.VolumeManager(size_u, size_v, size_w)

        # Set control points
        for u in range(size_u):
            for v in range(size_v):
                for w in range(size_w):
                    # 'pt' is the control point, e.g. [10, 15, 12]
                    points.set_ctrlpt(pt, u, v, w)

        # Create spline geometry
        volume = BSpline.Volume()

        # Set control points
        volume.ctrlpts = points.ctrlpts
    """
    def __init__(self, *args, **kwargs):
        super(VolumeManager, self).__init__(*args, **kwargs)

    def find_index(self, *args):
        super(VolumeManager, self).find_index(*args)
        return args[1] + (args[0] * self._size[1]) + (args[2] * self._size[0] * self._size[1])
